@@ -1,5 +1,4 @@
 /* Input history for input fields. */
-/* $Id: inphist.c,v 1.74 2004/01/05 11:29:40 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -13,10 +12,10 @@
 #include "bfu/dialog.h"
 #include "bfu/inphist.h"
 #include "bfu/menu.h"
+#include "config/home.h"
 #include "config/options.h"
-#include "lowlevel/home.h"
+#include "dialogs/menu.h"
 #include "terminal/terminal.h"
-#include "terminal/window.h"
 #include "util/conv.h"
 #include "util/file.h"
 #include "util/lists.h"
@@ -25,34 +24,43 @@
 
 
 static void
-tab_compl_n(struct terminal *term, unsigned char *item, int len,
-	    struct window *win)
+tab_compl_n(struct dialog_data *dlg_data, unsigned char *item, int len)
 {
-	struct term_event ev = INIT_TERM_EVENT(EV_REDRAW, term->width, term->height, 0);
-	struct dialog_data *dlg_data = (struct dialog_data *) win->data;
 	struct widget_data *widget_data = selected_widget(dlg_data);
 
+	assert(widget_is_textfield(widget_data));
+
 	int_upper_bound(&len, widget_data->widget->datalen - 1);
 	memcpy(widget_data->cdata, item, len);
 	widget_data->cdata[len] = 0;
 	widget_data->info.field.cpos = len;
 	widget_data->info.field.vpos = 0;
-	dialog_func(win, &ev, 0);
+
+	redraw_dialog(dlg_data, 1);
 }
 
-static inline void
-tab_compl(struct terminal *term, unsigned char *item, struct window *win)
+static void
+tab_compl(struct dialog_data *dlg_data, unsigned char *item)
 {
-	tab_compl_n(term, item, strlen(item), win);
+	tab_compl_n(dlg_data, item, strlen(item));
+}
+
+/* menu_func_T */
+static void
+menu_tab_compl(struct terminal *term, void *item_, void *dlg_data_)
+{
+	unsigned char *item = item_;
+	struct dialog_data *dlg_data = dlg_data_;
+
+	tab_compl_n(dlg_data, item, strlen(item));
 }
 
 /* Complete to last unambiguous character, and display menu for all possible
  * further completions. */
 void
-do_tab_compl(struct terminal *term, struct list_head *history,
-	     struct window *win)
+do_tab_compl(struct dialog_data *dlg_data, struct list_head *history)
 {
-	struct dialog_data *dlg_data = (struct dialog_data *) win->data;
+	struct terminal *term = dlg_data->win->term;
 	struct widget_data *widget_data = selected_widget(dlg_data);
 	int cdata_len = strlen(widget_data->cdata);
 	int n = 0;
@@ -65,74 +73,120 @@ do_tab_compl(struct terminal *term, struct list_head *history,
 		if (strncmp(widget_data->cdata, entry->data, cdata_len))
 			continue;
 
-		add_to_menu(&items, entry->data, NULL, ACT_NONE,
-			    (menu_func) tab_compl, entry->data, 0);
+		add_to_menu(&items, entry->data, NULL, ACT_MAIN_NONE,
+			    menu_tab_compl, entry->data, 0);
 		n++;
 	}
 
-	if (n) {
-		if (n == 1) {
-			tab_compl(term, items->data, win);
-			mem_free(items);
-			return;
-		}
-
-		do_menu_selected(term, items, win, n - 1, 0);
+	if (n > 1) {
+		do_menu_selected(term, items, dlg_data, n - 1, 0);
+	} else {
+		if (n == 1) tab_compl(dlg_data, items->data);
+		mem_free(items);
 	}
 }
 
+/* Return the length of the common substring from the starts
+ * of the two strings a and b. */
+static inline int
+strcommonlen(unsigned char *a, unsigned char *b)
+{
+	unsigned char *start = a;
+
+	while (*a && *a == *b)
+		++a, ++b;
+
+	return a - start;
+}
+
 /* Complete to the last unambiguous character. Eg., I've been to google.com,
  * google.com/search?q=foo, and google.com/search?q=bar.  This function then
  * completes `go' to `google.com' and `google.com/' to `google.com/search?q='.
  */
 void
-do_tab_compl_unambiguous(struct terminal *term, struct list_head *history,
-			 struct window *win)
+do_tab_compl_unambiguous(struct dialog_data *dlg_data, struct list_head *history)
 {
-	struct dialog_data *dlg_data = (struct dialog_data *) win->data;
 	struct widget_data *widget_data = selected_widget(dlg_data);
 	int base_len = strlen(widget_data->cdata);
 	/* Maximum number of characters in a match. Characters after this
-	 * position are varying in other matches. Zero means that no max has
-	 * been set yet. */
+	 * position are varying in other matches. */
 	int longest_common_match = 0;
 	unsigned char *match = NULL;
 	struct input_history_entry *entry;
 
 	foreach (entry, *history) {
 		unsigned char *cur = entry->data;
-		unsigned char *matchpos = match ? match : widget_data->cdata;
-		int cur_len = 0;
-
-		for (; *cur && *cur == *matchpos; ++cur, ++matchpos) {
-			++cur_len;
-
-			/* XXX: I think that unifying the two cases of this
-			 * test could seriously hurt readability. --pasky */
-			if (longest_common_match
-			    && cur_len >= longest_common_match)
-				break;
-		}
+		int cur_len = strcommonlen(cur, match ? match
+		                                      : widget_data->cdata);
 
+		/* Throw away it away if it isn't even as long as what the user
+		 * entered. */
 		if (cur_len < base_len)
 			continue;
 
-		if (!match) cur_len = strlen(entry->data);
-
-		/* By now, @cur_len oscillates between @base_len and
-		 * @longest_common_match. */
-		if (longest_common_match
-		    && cur_len >= longest_common_match)
-			continue;
-
-		/* We found the next shortest common match. */
-		longest_common_match = cur_len;
-		match = entry->data;
+		if (!match) {
+			/* This is the first match, so its length is the maximum
+			 * for any future matches. */
+			longest_common_match = strlen(entry->data);
+			match = entry->data;
+		} else if (cur_len < longest_common_match) {
+			/* The current match has a shorter substring in common
+			 * with the previous candidates, so the common substring
+			 * shrinks. */
+			longest_common_match = cur_len;
+		}
 	}
 
 	if (!match) return;
 
-	tab_compl_n(term, match, longest_common_match, win);
+	tab_compl_n(dlg_data, match, longest_common_match);
+}
+
+
+/* menu_func_T */
+static void
+set_complete_file_menu(struct terminal *term, void *filename_, void *dlg_data_)
+{
+	struct dialog_data *dlg_data = dlg_data_;
+	struct widget_data *widget_data = selected_widget(dlg_data);
+	unsigned char *filename = filename_;
+	int filenamelen;
+
+	assert(widget_is_textfield(widget_data));
+
+	filenamelen = int_min(widget_data->widget->datalen - 1, strlen(filename));
+	memcpy(widget_data->cdata, filename, filenamelen);
+
+	widget_data->cdata[filenamelen] = 0;
+	widget_data->info.field.cpos = filenamelen;
+	widget_data->info.field.vpos = 0;
+
+	mem_free(filename);
+
+	redraw_dialog(dlg_data, 1);
+}
+
+/* menu_func_T */
+static void
+tab_complete_file_menu(struct terminal *term, void *path_, void *dlg_data_)
+{
+	struct dialog_data *dlg_data = dlg_data_;
+	unsigned char *path = path_;
+
+	auto_complete_file(term, 0 /* no_elevator */, path,
+			   set_complete_file_menu, tab_complete_file_menu,
+			   dlg_data);
+}
+
+void
+do_tab_compl_file(struct dialog_data *dlg_data, struct list_head *history)
+{
+	struct widget_data *widget_data = selected_widget(dlg_data);
+
+	if (get_cmd_opt_bool("anonymous"))
+		return;
+
+	tab_complete_file_menu(dlg_data->win->term, widget_data->cdata, dlg_data);
 }
 
 
@@ -155,14 +209,13 @@ check_duplicate_entries(struct input_history *history, unsigned char *data)
 		duplicate = entry;
 		entry = entry->prev;
 
-		del_from_list(duplicate);
+		del_from_history_list(history, duplicate);
 
 		/* Save the first duplicate entry */
 		if (!first_duplicate) {
 			first_duplicate = duplicate;
 		} else {
 			mem_free(duplicate);
-			history->size--;
 		}
 	}
 
@@ -188,24 +241,23 @@ add_to_input_history(struct input_history *history, unsigned char *data,
 	if (check_duplicate) {
 		entry = check_duplicate_entries(history, data);
 		if (entry) {
-			add_to_list(history->entries, entry);
-			if (!history->nosave) history->dirty = 1;
+			add_to_history_list(history, entry);
 			return;
 		}
 	}
 
 	/* Copy it all etc. */
 	/* One byte is already reserved for url in struct input_history_item. */
-	entry = mem_alloc(sizeof(struct input_history_entry) + length);
+	entry = mem_alloc(sizeof(*entry) + length);
 	if (!entry) return;
 
 	memcpy(entry->data, data, length + 1);
 
 	add_to_history_list(history, entry);
 
-	/* limit size of history to MAX_INPUT_HISTORY_ENTRIES
-	 * removing first entries if needed */
-	for (; history->size > MAX_INPUT_HISTORY_ENTRIES; history->size--) {
+	/* Limit size of history to MAX_INPUT_HISTORY_ENTRIES, removing first
+	 * entries if needed. */
+	while (history->size > MAX_INPUT_HISTORY_ENTRIES) {
 		if (list_empty(history->entries)) {
 			INTERNAL("history is empty");
 			history->size = 0;
@@ -213,7 +265,7 @@ add_to_input_history(struct input_history *history, unsigned char *data,
 		}
 
 		entry = history->entries.prev;
-		del_from_list(entry);
+		del_from_history_list(history, entry);
 		mem_free(entry);
 	}
 }
@@ -226,27 +278,28 @@ load_input_history(struct input_history *history, unsigned char *filename)
 	unsigned char line[MAX_STR_LEN];
 	FILE *file;
 
-	if (get_opt_int_tree(cmdline_options, "anonymous")) return 0;
+	if (get_cmd_opt_bool("anonymous")) return 0;
 	if (elinks_home) {
 		history_file = straconcat(elinks_home, filename, NULL);
 		if (!history_file) return 0;
 	}
 
-	history->nosave = 1;
-
-	file = fopen(history_file, "r");
+	file = fopen(history_file, "rb");
 	if (elinks_home) mem_free(history_file);
 	if (!file) return 0;
 
-	while (safe_fgets(line, MAX_STR_LEN, file)) {
+	history->nosave = 1;
+
+	while (fgets(line, MAX_STR_LEN, file)) {
 		/* Drop '\n'. */
 		if (*line) line[strlen(line) - 1] = 0;
 		add_to_input_history(history, line, 0);
 	}
 
-	fclose(file);
 	history->nosave = 0;
 
+	fclose(file);
+
 	return 0;
 }
 
@@ -262,7 +315,7 @@ save_input_history(struct input_history *history, unsigned char *filename)
 
 	if (!history->dirty
 	    || !elinks_home
-	    || get_opt_int_tree(cmdline_options, "anonymous"))
+	    || get_cmd_opt_bool("anonymous"))
 		return 0;
 
 	history_file = straconcat(elinks_home, filename, NULL);
@@ -283,3 +336,23 @@ save_input_history(struct input_history *history, unsigned char *filename)
 
 	return secure_close(ssi);
 }
+
+void
+dlg_set_history(struct widget_data *widget_data)
+{
+	assert(widget_has_history(widget_data));
+	assert(widget_data->widget->datalen > 0);
+
+	if ((void *) widget_data->info.field.cur_hist != &widget_data->info.field.history) {
+		unsigned char *s = widget_data->info.field.cur_hist->data;
+
+		widget_data->info.field.cpos = int_min(strlen(s), widget_data->widget->datalen - 1);
+		if (widget_data->info.field.cpos)
+			memcpy(widget_data->cdata, s, widget_data->info.field.cpos);
+	} else {
+		widget_data->info.field.cpos = 0;
+	}
+
+	widget_data->cdata[widget_data->info.field.cpos] = 0;
+	widget_data->info.field.vpos = int_max(0, widget_data->info.field.cpos - widget_data->box.width);
+}
