@@ -1,5 +1,4 @@
 /* RFC1524 (mailcap file) implementation */
-/* $Id: mailcap.c,v 1.87.2.4 2005/05/01 22:03:23 jonas Exp $ */
 
 /* This file contains various functions for implementing a fair subset of
  * rfc1524.
@@ -29,12 +28,12 @@
 
 #include "config/options.h"
 #include "intl/gettext/libintl.h"
+#include "main/module.h"
 #include "mime/backend/common.h"
 #include "mime/backend/mailcap.h"
 #include "mime/mime.h"
-#include "modules/module.h"
 #include "osdep/osdep.h"		/* For exe() */
-#include "sched/session.h"
+#include "session/session.h"
 #include "util/file.h"
 #include "util/hash.h"
 #include "util/lists.h"
