@@ -1,5 +1,4 @@
 /* Terminal screen drawing routines. */
-/* $Id: screen.c,v 1.126 2004/01/04 00:07:50 zas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -11,6 +10,7 @@
 #include "elinks.h"
 
 #include "config/options.h"
+#include "intl/charsets.h"
 #include "osdep/ascii.h"
 #include "osdep/osdep.h"
 #include "terminal/color.h"
@@ -117,7 +117,7 @@ struct screen_driver {
 	enum color_mode color_mode;
 
 	/* These are directly derived from the terminal options. */
-	unsigned int trans:1;
+	unsigned int transparent:1;
 
 	/* The terminal._template_ name. */
 	unsigned char name[1]; /* XXX: Keep last! */
@@ -131,18 +131,18 @@ static struct screen_driver dumb_screen_driver = {
 	/* frame_seqs: */	NULL,
 	/* underline: */	underline_seqs,
 	/* color_mode: */	COLOR_MODE_16,
-	/* trans: */		1,
+	/* transparent: */	1,
 };
 
 static struct screen_driver vt100_screen_driver = {
 				NULL_LIST_HEAD,
 	/* type: */		TERM_VT100,
 	/* charsets: */		{ -1, -1 },	/* No UTF8 I/O */
-	/* frame: */		frame_vt100,	/* No UTF8 I/O */
-	/* frame_seqs: */	vt100_frame_seqs, /* No UTF8 I/O */
+	/* frame: */		frame_vt100,
+	/* frame_seqs: */	vt100_frame_seqs,
 	/* underline: */	underline_seqs,
 	/* color_mode: */	COLOR_MODE_16,
-	/* trans: */		1,
+	/* transparent: */	1,
 };
 
 static struct screen_driver linux_screen_driver = {
@@ -153,7 +153,7 @@ static struct screen_driver linux_screen_driver = {
 	/* frame_seqs: */	NULL,		/* No m11_hack */
 	/* underline: */	underline_seqs,
 	/* color_mode: */	COLOR_MODE_16,
-	/* trans: */		1,
+	/* transparent: */	1,
 };
 
 static struct screen_driver koi8_screen_driver = {
@@ -164,7 +164,7 @@ static struct screen_driver koi8_screen_driver = {
 	/* frame_seqs: */	NULL,
 	/* underline: */	underline_seqs,
 	/* color_mode: */	COLOR_MODE_16,
-	/* trans: */		1,
+	/* transparent: */	1,
 };
 
 static struct screen_driver freebsd_screen_driver = {
@@ -175,7 +175,7 @@ static struct screen_driver freebsd_screen_driver = {
 	/* frame_seqs: */	NULL,		/* No m11_hack */
 	/* underline: */	underline_seqs,
 	/* color_mode: */	COLOR_MODE_16,
-	/* trans: */		1,
+	/* transparent: */	1,
 };
 
 /* XXX: Keep in sync with enum term_mode_type. */
@@ -187,6 +187,7 @@ static struct screen_driver *screen_drivers[] = {
 	/* TERM_FREEBSD: */	&freebsd_screen_driver,
 };
 
+
 static INIT_LIST_HEAD(active_screen_drivers);
 
 static void
@@ -195,7 +196,7 @@ update_screen_driver(struct screen_driver *driver, struct option *term_spec)
 	int utf8_io = get_opt_bool_tree(term_spec, "utf_8_io");
 
 	driver->color_mode = get_opt_int_tree(term_spec, "colors");
-	driver->trans = get_opt_bool_tree(term_spec, "transparency");
+	driver->transparent = get_opt_bool_tree(term_spec, "transparency");
 
 	if (get_opt_bool_tree(term_spec, "underline")) {
 		driver->underline = underline_seqs;
@@ -205,47 +206,41 @@ update_screen_driver(struct screen_driver *driver, struct option *term_spec)
 
 	if (utf8_io) {
 		driver->charsets[0] = get_opt_int_tree(term_spec, "charset");
-	} else {
-		driver->charsets[0] = -1;
-	}
+		if (driver->type == TERM_LINUX) {
+			if (get_opt_bool_tree(term_spec, "restrict_852"))
+				driver->frame = frame_restrict;
 
-	if (driver->type == TERM_LINUX) {
-		if (get_opt_bool_tree(term_spec, "restrict_852")) {
-			driver->frame = frame_restrict;
-		}
-
-		if (utf8_io) {
 			driver->charsets[1] = get_cp_index("cp437");
 
-		} else if (get_opt_bool_tree(term_spec, "m11_hack")) {
-			driver->frame_seqs = m11_hack_frame_seqs;
-		}
-
-	} else if (driver->type == TERM_FREEBSD) {
-		if (utf8_io) {
+		} else if (driver->type == TERM_FREEBSD) {
 			driver->charsets[1] = get_cp_index("cp437");
 
-		} else if (get_opt_bool_tree(term_spec, "m11_hack")) {
-			driver->frame_seqs = m11_hack_frame_seqs;
-		}
-
-
-	} else if (driver->type == TERM_VT100) {
-		if (utf8_io) {
+		} else if (driver->type == TERM_VT100) {
 			driver->frame = frame_vt100_u;
 			driver->charsets[1] = get_cp_index("cp437");
-		} else {
-			driver->frame = frame_vt100;
-		}
 
-	} else if (driver->type == TERM_KOI8) {
-		if (utf8_io) {
+		} else if (driver->type == TERM_KOI8) {
 			driver->charsets[1] = get_cp_index("koi8-r");
+
+		} else {
+			driver->charsets[1] = driver->charsets[0];
 		}
 
 	} else {
-		if (utf8_io) {
-			driver->charsets[1] = driver->charsets[0];
+		driver->charsets[0] = -1;
+		if (driver->type == TERM_LINUX) {
+			if (get_opt_bool_tree(term_spec, "restrict_852"))
+				driver->frame = frame_restrict;
+
+			if (get_opt_bool_tree(term_spec, "m11_hack"))
+				driver->frame_seqs = m11_hack_frame_seqs;
+
+		} else if (driver->type == TERM_FREEBSD) {
+			if (get_opt_bool_tree(term_spec, "m11_hack"))
+				driver->frame_seqs = m11_hack_frame_seqs;
+
+		} else if (driver->type == TERM_VT100) {
+			driver->frame = frame_vt100;
 		}
 	}
 }
@@ -274,11 +269,11 @@ add_screen_driver(enum term_mode_type type, struct terminal *term, int env_len)
 	struct screen_driver *driver;
 
 	/* One byte is reserved for name in struct screen_driver. */
-	driver = mem_alloc(sizeof(struct screen_driver) + env_len);
+	driver = mem_alloc(sizeof(*driver) + env_len);
 	if (!driver) return NULL;
 
-	memcpy(driver, screen_drivers[type], sizeof(struct screen_driver) - 1);
-	memcpy(driver->name, term->term, env_len + 1);
+	memcpy(driver, screen_drivers[type], sizeof(*driver) - 1);
+	memcpy(driver->name, term->spec->name, env_len + 1);
 
 	add_to_list(active_screen_drivers, driver);
 
@@ -298,15 +293,13 @@ get_screen_driver(struct terminal *term)
 	struct screen_driver *driver;
 
 	foreach (driver, active_screen_drivers) {
-		if (driver->type == type && !memcmp(driver->name, name, len)) {
-			/* Some simple probably useless MRU ;) */
-			if (driver != active_screen_drivers.next) {
-				del_from_list(driver);
-				add_to_list(active_screen_drivers, driver);
-			}
+		if (driver->type != type) continue;
+		if (memcmp(driver->name, name, len + 1)) continue;
 
-			return driver;
-		}
+		/* Some simple probably useless MRU ;) */
+		move_to_top_of_list(active_screen_drivers, driver);
+
+		return driver;
 	}
 
 	return add_screen_driver(type, term, len);
@@ -324,26 +317,25 @@ done_screen_drivers(void)
 static inline struct string *
 add_cursor_move_to_string(struct string *screen, int y, int x)
 {
-	/* 28 chars for both of the @y and @x numbers should be enough. */
-	unsigned char code[32];
-	int length = 2;
-	int ret;
+#define CURSOR_NUM_LEN 10 /* 10 chars for @y and @x numbers should be more than enough. */
+	unsigned char code[4 + 2 * CURSOR_NUM_LEN + 1];
+	unsigned int length = 2;
 
 	code[0] = '\033';
 	code[1] = '[';
 
-	ret = longcat(code, &length, y, 30, 0);
-	/* Make sure theres atleast room for ';' and `some' number ;) */
-	if (ret < 0 || length > 30) return NULL;
+	if (ulongcat(code, &length, y, CURSOR_NUM_LEN, 0) < 0)
+		return screen;
 
 	code[length++] = ';';
 
-	ret = longcat(code, &length, x, sizeof(code) - length, 0);
-	if (ret < 0 || length > 31) return NULL;
+	if (ulongcat(code, &length, x, CURSOR_NUM_LEN, 0) < 0)
+		return screen;
 
 	code[length++] = 'H';
 
 	return add_bytes_to_string(screen, code, length);
+#undef CURSOR_NUM_LEN
 }
 
 struct screen_state {
@@ -351,15 +343,22 @@ struct screen_state {
 	unsigned char underline;
 	unsigned char bold;
 	unsigned char attr;
+	/* Following should match struct screen_char color field. */
+#if defined(CONFIG_88_COLORS) || defined(CONFIG_256_COLORS)
 	unsigned char color[2];
+#else
+	unsigned char color[1];
+#endif
 };
 
-#ifdef CONFIG_256_COLORS
-#define compare_color(a, b)	(!memcmp((a), (b), 2))
-#define copy_color(a, b)	memcpy((a), (b), 2)
+#if defined(CONFIG_88_COLORS) || defined(CONFIG_256_COLORS)
+#define compare_color(a, b)	((a)[0] == (b)[0] && (a)[1] == (b)[1])
+#define copy_color(a, b)	do { (a)[0] = (b)[0]; (a)[1] = (b)[1]; } while (0)
+#define INIT_SCREEN_STATE 	{ 0xFF, 0xFF, 0xFF, 0, { 0xFF, 0xFF } }
 #else
 #define compare_color(a, b)	((a)[0] == (b)[0])
-#define copy_color(a, b)	((a)[0] = (b)[0])
+#define copy_color(a, b)	do { (a)[0] = (b)[0]; } while (0)
+#define INIT_SCREEN_STATE 	{ 0xFF, 0xFF, 0xFF, 0, { 0xFF } }
 #endif
 
 #define compare_bg_color(a, b)	(TERM_COLOR_BACKGROUND(a) == TERM_COLOR_BACKGROUND(b))
@@ -371,21 +370,22 @@ static inline void
 add_char_data(struct string *screen, struct screen_driver *driver,
 	      unsigned char data, unsigned char border)
 {
-	if (data >= ' ' && data != ASCII_DEL /* && c != 155*/) {
-		if (border && driver->frame && data >= 176 && data < 224) {
-			data = driver->frame[data - 176];
-		}
-
-		if (use_utf8_io(driver)) {
-			int charset = driver->charsets[!!border];
-
-			add_to_string(screen, cp2utf_8(charset, data));
-		} else {
-			add_char_to_string(screen, data);
-		}
-	} else {
+	if (!isscreensafe(data)) {
 		add_char_to_string(screen, ' ');
+		return;
 	}
+
+	if (border && driver->frame && data >= 176 && data < 224)
+		data = driver->frame[data - 176];
+
+	if (use_utf8_io(driver)) {
+		int charset = driver->charsets[!!border];
+
+		add_to_string(screen, cp2utf_8(charset, data));
+		return;
+	}
+
+	add_char_to_string(screen, data);
 }
 
 /* Time critical section. */
@@ -423,13 +423,13 @@ add_char16(struct string *screen, struct screen_driver *driver,
 		add_bytes_to_string(screen, "\033[0", 3);
 
 		if (driver->color_mode == COLOR_MODE_16) {
-			static unsigned char code[6] = ";30;40";
+			unsigned char code[6] = ";30;40";
 			unsigned char bgcolor = TERM_COLOR_BACKGROUND(ch->color);
 
-			code[2] = '0' + TERM_COLOR_FOREGROUND(ch->color);
+			code[2] += TERM_COLOR_FOREGROUND(ch->color);
 
-			if (!driver->trans || bgcolor != 0) {
-				code[5] = '0' + bgcolor;
+			if (!driver->transparent || bgcolor != 0) {
+				code[5] += bgcolor;
 				add_bytes_to_string(screen, code, 6);
 			} else {
 				add_bytes_to_string(screen, code, 3);
@@ -446,7 +446,7 @@ add_char16(struct string *screen, struct screen_driver *driver,
 		}
 
 		/* Check if the char should be rendered bold. */
-		if (ch->attr & SCREEN_ATTR_BOLD) {
+		if (bold) {
 			add_bytes_to_string(screen, ";1", 2);
 		}
 
@@ -456,7 +456,7 @@ add_char16(struct string *screen, struct screen_driver *driver,
 	add_char_data(screen, driver, ch->data, border);
 }
 
-#ifdef CONFIG_256_COLORS
+#if defined(CONFIG_88_COLORS) || defined(CONFIG_256_COLORS)
 static struct string color256_seqs[] = {
 	/* foreground: */	TERM_STRING("\033[0;38;5;%dm"),
 	/* background: */	TERM_STRING("\033[48;5;%dm"),
@@ -466,39 +466,39 @@ static inline void
 add_char_color(struct string *screen, struct string *seq, unsigned char color)
 {
 	unsigned char color_buf[3];
-	unsigned char *color_pos;
-	int seq_pos, color_len;
+	unsigned char *color_pos = color_buf;
+	int seq_pos = 0;
+       	int color_len = 1;
 
 	check_string_magic(seq);
-	for (seq_pos = 0; seq->source[seq_pos] != '%'; seq_pos++) ;
+	for (; seq->source[seq_pos] != '%'; seq_pos++) ;
 
 	add_bytes_to_string(screen, seq->source, seq_pos);
 
-	if (color > 9) {
+	if (color < 10) {
+		color_pos += 2;
+	} else {
 		int color2;
 
-		if (color > 199) {
-			color_buf[0] = '2';
-			color_pos = color_buf;
-			color_len = 3;
-			color -= 200;
-		} else if (color > 99) {
-			color_buf[0] = '1';
-			color_pos = color_buf;
-			color_len = 3;
-			color -= 100;
+		++color_len;
+		if (color < 100) {
+			++color_pos;
 		} else {
-			color_len = 2;
-			color_pos = color_buf + 1;
+			++color_len;
+
+			if (color < 200) {
+				color_buf[0] = '1';
+				color -= 100;
+			} else {
+				color_buf[0] = '2';
+				color -= 200;
+			}
 		}
 
 		color2 = (color % 10);
 		color /= 10;
 		color_buf[1] = '0' + color;
 		color = color2;
-	} else {
-		color_len = 1;
-		color_pos = color_buf + 2;
 	}
 
 	color_buf[2] = '0' + color;
@@ -525,12 +525,16 @@ add_char256(struct string *screen, struct screen_driver *driver,
 			add_term_string(screen, driver->frame_seqs[state->border]);
 		}
 
+		if ((attr_delta & SCREEN_ATTR_UNDERLINE) && driver->underline) {
+			state->underline = !!(ch->attr & SCREEN_ATTR_UNDERLINE);
+			add_term_string(screen, driver->underline[state->underline]);
+		}
+
 		if (attr_delta & SCREEN_ATTR_BOLD) {
 			if (ch->attr & SCREEN_ATTR_BOLD) {
 				add_bytes_to_string(screen, "\033[1m", 4);
 			} else {
 				/* Force repainting of the other attributes. */
-				attr_delta |= SCREEN_ATTR_UNDERLINE;
 				state->color[0] = ch->color[0] + 1;
 			}
 		}
@@ -542,17 +546,17 @@ add_char256(struct string *screen, struct screen_driver *driver,
 		copy_color(state->color, ch->color);
 
 		add_foreground_color(screen, color256_seqs, ch);
-		if (!driver->trans || ch->color[1] != 0) {
+		if (!driver->transparent || ch->color[1] != 0) {
 			add_background_color(screen, color256_seqs, ch);
 		}
 
 		if (ch->attr & SCREEN_ATTR_BOLD)
 			add_bytes_to_string(screen, "\033[1m", 4);
-	}
 
-	if ((attr_delta & SCREEN_ATTR_UNDERLINE) && driver->underline) {
-		state->underline = !!(ch->attr & SCREEN_ATTR_UNDERLINE);
-		add_term_string(screen, driver->underline[state->underline]);
+		if (ch->attr & SCREEN_ATTR_UNDERLINE && driver->underline) {
+			state->underline = !!(ch->attr & SCREEN_ATTR_UNDERLINE);
+			add_term_string(screen, driver->underline[state->underline]);
+		}
 	}
 
 	add_char_data(screen, driver, ch->data, ch->attr & SCREEN_ATTR_FRAME);
@@ -561,20 +565,35 @@ add_char256(struct string *screen, struct screen_driver *driver,
 
 #define add_chars(image_, term_, driver_, state_, ADD_CHAR)			\
 {										\
-	register int y = (term_)->screen->dirty_from;				\
-	int ypos = y * term->width;						\
-	register struct screen_char *current = &(term_)->screen->last_image[ypos]; \
-	register struct screen_char *pos = &(term_)->screen->image[ypos];	\
-	register struct screen_char *prev_pos = NULL;				\
+	struct terminal_screen *screen = (term_)->screen;			\
+	int y = screen->dirty_from;					\
+	int ypos = y * (term_)->width;						\
 	int prev_y = -1;							\
+	int xmax = (term_)->width - 1;						\
+	int ymax = (term_)->height - 1;						\
+	struct screen_char *current = &screen->last_image[ypos];	\
+	struct screen_char *pos = &screen->image[ypos];		\
+	struct screen_char *prev_pos = NULL; /* Warning prevention. */	\
 										\
-	if ((term_)->screen->dirty_to >= (term_)->height)			\
-		(term_)->screen->dirty_to = (term_)->height - 1;		\
+	int_upper_bound(&screen->dirty_to, ymax);				\
 										\
-	for (; y <= (term_)->screen->dirty_to; y++) {				\
-		register int x = 0;						\
+	for (; y <= screen->dirty_to; y++) {					\
+		int is_last_line = (y == ymax);					\
+		int x = 0;						\
+										\
+		for (; x <= xmax; x++, current++, pos++) {			\
+			/*  Workaround for terminals without
+			 *  "eat_newline_glitch (xn)", e.g., the cons25 family
+			 *  of terminals and cygwin terminal.
+			 *  It prevents display distortion, but char at bottom
+			 *  right of terminal will not be drawn.
+			 *  A better fix would be to correctly detects
+			 *  terminal type, and/or add a terminal option for
+			 *  this purpose. */					\
+										\
+			if (is_last_line && x == xmax)				\
+				break;						\
 										\
-		for (; x < (term_)->width; x++, current++, pos++) {		\
 			if (compare_bg_color(pos->color, current->color)) {	\
 				/* No update for exact match. */		\
 				if (compare_fg_color(pos->color, current->color)\
@@ -596,7 +615,7 @@ add_char256(struct string *screen, struct screen_driver *driver,
 				prev_pos = pos;					\
 				prev_y = y;					\
 			}							\
-										\
+				 						\
 			for (; prev_pos <= pos ; prev_pos++)			\
 				ADD_CHAR(image_, driver_, prev_pos, state_);	\
 		}								\
@@ -608,32 +627,39 @@ add_char256(struct string *screen, struct screen_driver *driver,
 void
 redraw_screen(struct terminal *term)
 {
-	struct screen_driver *driver = get_screen_driver(term);
+	struct screen_driver *driver;
 	struct string image;
-	struct screen_state state = { 0xFF, 0xFF, 0xFF, 0, { 0xFF, 0xFF } };
+	struct screen_state state = INIT_SCREEN_STATE;
 	struct terminal_screen *screen = term->screen;
 
-	if (!driver
-	    || !screen
-	    || (screen->dirty_from > screen->dirty_to)
-	    || (term->master && is_blocked())
-	    || !init_string(&image)) return;
+	if (!screen || screen->dirty_from > screen->dirty_to) return;
+	if (term->master && is_blocked()) return;
+
+	driver = get_screen_driver(term);
+	if (!driver) return;
+
+	if (!init_string(&image)) return;
 
 	switch (driver->color_mode) {
-	case COLOR_MODES:
-	case COLOR_MODE_DUMP:
-		INTERNAL("Invalid color mode");
-		return;
-
-	case COLOR_MODE_256:
-#ifdef CONFIG_256_COLORS
-		add_chars(&image, term, driver, &state, add_char256);
-		break;
-#endif
 	case COLOR_MODE_MONO:
 	case COLOR_MODE_16:
 		add_chars(&image, term, driver, &state, add_char16);
 		break;
+#ifdef CONFIG_88_COLORS
+	case COLOR_MODE_88:
+		add_chars(&image, term, driver, &state, add_char256);
+		break;
+#endif
+#ifdef CONFIG_256_COLORS
+	case COLOR_MODE_256:
+		add_chars(&image, term, driver, &state, add_char256);
+		break;
+#endif
+	case COLOR_MODES:
+	case COLOR_MODE_DUMP:
+	default:
+		INTERNAL("Invalid color mode (%d).", driver->color_mode);
+		return;
 	}
 
 	if (image.length) {
@@ -643,18 +669,21 @@ redraw_screen(struct terminal *term)
 		add_bytes_to_string(&image, "\033[0m", 4);
 
 		/* If we ended in border state end the frame mode. */
-		if (state.border && driver->frame_seqs) {
+		if (state.border && driver->frame_seqs)
 			add_term_string(&image, driver->frame_seqs[0]);
-		}
+
 	}
 
+	/* Even if nothing was redrawn, we possibly still need to move
+	 * cursor. */
 	if (image.length
 	    || screen->cx != screen->lcx
 	    || screen->cy != screen->lcy) {
 		screen->lcx = screen->cx;
 		screen->lcy = screen->cy;
 
-		add_cursor_move_to_string(&image, screen->cy + 1, screen->cx + 1);
+		add_cursor_move_to_string(&image, screen->cy + 1,
+						  screen->cx + 1);
 	}
 
 	if (image.length) {
@@ -685,7 +714,11 @@ erase_screen(struct terminal *term)
 void
 beep_terminal(struct terminal *term)
 {
+#ifdef CONFIG_WIN32
+	MessageBeep(MB_ICONEXCLAMATION);
+#else
 	hard_write(term->fdout, "\a", 1);
+#endif
 }
 
 struct terminal_screen *
@@ -693,7 +726,7 @@ init_screen(void)
 {
 	struct terminal_screen *screen;
 
-	screen = mem_calloc(1, sizeof(struct terminal_screen));
+	screen = mem_calloc(1, sizeof(*screen));
 	if (!screen) return NULL;
 
 	screen->lcx = -1;
@@ -721,7 +754,7 @@ resize_screen(struct terminal *term, int width, int height)
 	size = width * height;
 	if (size <= 0) return;
 
-	bsize = size * sizeof(struct screen_char);
+	bsize = size * sizeof(*image);
 
 	image = mem_realloc(screen->image, bsize * 2);
 	if (!image) return;
@@ -740,6 +773,6 @@ resize_screen(struct terminal *term, int width, int height)
 void
 done_screen(struct terminal_screen *screen)
 {
-	if (screen->image) mem_free(screen->image);
+	mem_free_if(screen->image);
 	mem_free(screen);
 }
